%{
  import java.io.* ;
%}

/* Terminals (tokens returned by the scanner). */

%token S, K, I, Dieze ;
%token Arobase ;
%token INTEGER ;
%token SYMBOL ;
%token LET , EGAL , DOUBLE_POINT_VIRGULE ;
%token UNDEF ; 
%token LEFT_BRACKET , COMMA, DOT , RIGHT_BRACKET ;
%token LEFT_PAREN , RIGHT_PAREN ;
%token LAMBDA_PLUS, LAMBDA_PLUS_PLUS, LAMBDA_PLUS_PLUS_PLUS, LAMBDA_PLUS_PLUS_PLUS_PLUS ;

%%

/* The grammar */

to_read:          
definitions
{ 
  $$ = nullSKParserVal ;
}
| expression
{ 
  expression = (SKTerm)($1.obj) ;
  $$ = nullSKParserVal ;
}
;

expression:
term_sequence  DOUBLE_POINT_VIRGULE
{ 
  $$ = new SKParserVal(SKBuilder.buildApplication((SKSequence)($1.obj))) ; 
}
;

definitions:
/* empty */
{ 
  $$ = nullSKParserVal ;
}
| definition definitions 
{ 
  $$ = nullSKParserVal ;
}
;

definition:
LET symbol EGAL term_sequence DOUBLE_POINT_VIRGULE 
{
  SKBuilder.registerDefinition((SKSymbol)($2.obj),
			       SKBuilder.buildApplication((SKSequence)($4.obj))) ; 
  $$ = new SKParserVal((SKTerm)null) ; 

  if (DEBUG) {
    try {
      writer.write("Rule: LET symbol EGAL term_sequence DOUBLE_POINT_VIRGULE\nwith symbol = ") ;
      ((SKSymbol)($2.obj)).writeTo(writer) ;
      writer.write("\nand term_sequence = ") ;
      ((SKSequence)($4.obj)).writeTo(writer) ;
      writer.write("\nresult = ") ;
      (SKBuilder.buildApplication((SKSequence)($4.obj))).writeTo(writer) ;
      writer.write("\n\n") ;
      writer.flush() ;
    } catch (Exception e) { e.printStackTrace() ; }
  }
}
| UNDEF symbol DOUBLE_POINT_VIRGULE
{
  SKBuilder.registerDefinition((SKSymbol)($2.obj),null) ;
  $$ = new SKParserVal((SKTerm)null) ; 
}
;

symbol:
SYMBOL
{
  $$ = new SKParserVal(SKBuilder.getSymbol(skLexer.yytext())) ;
}

term_sequence:
term
{ 
  $$ = new SKParserVal(new SKSequence((SKTerm)($1.obj))) ;
}
| term term_sequence 
{ 
  ((SKSequence)($2.obj)).add((SKTerm)($1.obj)) ;
  $$ = $2 ; 
}
;

term:             
S                                     
{ 
  $$ = new SKParserVal(CombinatorS.unique) ;  
}
| K
{ 
  $$ = new SKParserVal(CombinatorK.unique) ;  
}
| I
{ 
  $$ = new SKParserVal(CombinatorI.unique) ;  
}
| Dieze
{ 
  $$ = new SKParserVal(CombinatorDieze.unique) ;  
}
| symbol
{ 
  $$ = $1 ;
}
| INTEGER                               
{ 
  $$ = new SKParserVal(SKBuilder.getNumber(Integer.parseInt(skLexer.yytext()))) ;  
}
| LEFT_BRACKET inner_list RIGHT_BRACKET 
{ 
  $$ = $2 ; 
}		
| LEFT_PAREN term_sequence RIGHT_PAREN
{ 
  $$ = new SKParserVal(SKBuilder.buildApplication((SKSequence)($2.obj))) ; 
}
| LEFT_PAREN LAMBDA_PLUS symbol DOT term_sequence RIGHT_PAREN
{
  $$ = new SKParserVal(SKBuilder.lambdaPlus((SKSymbol)($3.obj),SKBuilder.buildApplication((SKSequence)($5.obj)))) ;
}
| LEFT_PAREN LAMBDA_PLUS_PLUS symbol DOT term_sequence RIGHT_PAREN
{
  $$ = new SKParserVal(SKBuilder.lambdaPlusPlus((SKSymbol)($3.obj),SKBuilder.buildApplication((SKSequence)($5.obj)))) ;
}
| LEFT_PAREN LAMBDA_PLUS_PLUS_PLUS symbol DOT term_sequence RIGHT_PAREN
{
  $$ = new SKParserVal(SKBuilder.lambdaPlusPlusPlus((SKSymbol)($3.obj),SKBuilder.buildApplication((SKSequence)($5.obj)))) ;
}
| LEFT_PAREN LAMBDA_PLUS_PLUS_PLUS_PLUS symbol DOT term_sequence RIGHT_PAREN
{
  $$ = new SKParserVal(SKBuilder.lambdaPlusPlusPlusPlus((SKSymbol)($3.obj),SKBuilder.buildApplication((SKSequence)($5.obj)))) ;
}
| Arobase symbol
{
  $$ = new SKParserVal(SKBuilder.buildArobase((SKSymbol)($2.obj))) ;
}
;

inner_list:       
term DOT term_sequence
{ 
  $$ = new SKParserVal(SKBuilder.getCons((SKTerm)($1.obj),SKBuilder.buildApplication((SKSequence)($3.obj)))) ;
}
| term COMMA inner_list 
{
  $$ = new SKParserVal(SKBuilder.getCons((SKTerm)($1.obj),(SKTerm)($3.obj))) ;
}
| term
{ 
  $$ = new SKParserVal(SKBuilder.getCons((SKTerm)($1.obj))) ; 
} 
;


	
%%

private static final boolean DEBUG = false ;

private Writer writer ;

public SKTerm expression = null ;

/* the "bad" value */
private SKParserVal nullSKParserVal = new SKParserVal((SKTerm)null) ;
  
/* a reference to the lexer object */
private SKLexer skLexer;

/* interface to the lexer */
private int yylex () 
{
    int yyl_return = -1;

    try {
      yylval = nullSKParserVal ;
      yyl_return = skLexer.yylex();
    }
    catch (Exception e) {}
    catch (Throwable e) {}

    return yyl_return;
}

/* error reporting */

public int errorLine = -1 ;

public void yyerror (String error) 
{
  errorLine = skLexer.getLine() ;
}

/* lexer is created in the constructor */
public SKParser(Reader r) 
{
  if (DEBUG) {
    try {
      writer = new BufferedWriter(new FileWriter("ski.dbg")) ;
    } catch (Exception e) { e.printStackTrace() ; }
  }

  skLexer = new SKLexer(r,this);
}

