import java.io.* ;

public class SKApplication extends SKTerm
{
    public SKTerm function ;
    public SKTerm argument ;

    public SKApplication(SKTerm function, SKTerm argument)
    {
	this.function = function ;
	this.argument = argument ;
    }

    private boolean arobaseEliminated = false ;

    public SKTerm elimArobases()
    {
	if (arobaseEliminated)
	    return this ;

	this.function = function.elimArobases() ;
	this.argument = argument.elimArobases() ;

	arobaseEliminated = true ;
	return this ;
    }

    public final boolean isAtom()
    {
	return false ;
    }

    public final boolean isApplication()
    {
	return true ;
    }

    public final SKTerm abstractPlus(SKSymbol x)
    {
	return SKBuilder.app(SKBuilder.S,
			     function.abstractPlus(x),
			     argument.abstractPlus(x)) ;
    }

    public final SKTerm abstractPlusPlus(SKSymbol x)
    {
	if (argument == x && function.freeOf(x))
	    return function ;

	return SKBuilder.app(SKBuilder.S,
			     function.abstractPlusPlus(x),
			     argument.abstractPlusPlus(x)) ;
    }

    public final SKTerm abstractPlusPlusPlus(SKSymbol x)
    {
	if (function.freeOf(x)) {

	    if (argument == x)
		return function ;

	    if (argument.freeOf(x))
		return SKBuilder.app(SKBuilder.K , this) ;
	}

	return SKBuilder.app(SKBuilder.S, function.abstractPlusPlusPlus(x), argument.abstractPlusPlusPlus(x)) ;
    }

    public final SKTerm abstractPlusPlusPlusPlus(SKSymbol x)
    {
	if (function.freeOf(x)) {

	    if (argument == x)
		return function ;

	    if (argument.isApplication()) {

		SKApplication a = (SKApplication)argument ;

		if (a.function.freeOf(x) && a.argument == x)
		    return SKBuilder.app( SKBuilder.getSymbol("B") , function , a.function ) ;
	    }

	    if (argument.freeOf(x))
		return new SKApplication(SKBuilder.K,this) ;

	    return SKBuilder.app( SKBuilder.getSymbol("B"), function, argument.abstractPlusPlusPlusPlus(x)) ;
	}

	if (function.isApplication()) {

	    SKApplication f = (SKApplication)function ;

	    if (f.function.freeOf(x) && f.argument == x) {

		if (argument == x)
		    return SKBuilder.app( SKBuilder.getSymbol("W") , f.function ) ;

		if (argument.isApplication()) {

		    SKApplication a = (SKApplication)argument ;

		    if (a.function.freeOf(x) && a.argument == x)
			return SKBuilder.app( SKBuilder.S , f.function , a.function ) ;
		}

		if (argument.freeOf(x))
		    return SKBuilder.app( SKBuilder.getSymbol("C") , f.function , argument ) ;

		return SKBuilder.app( SKBuilder.S , f.function , argument.abstractPlusPlusPlusPlus(x) ) ; 
	    }
	}

	if (function == x) {
	    if (argument.freeOf(x))
		return SKBuilder.app( SKBuilder.getSymbol("C*") , argument ) ;
	}

	return SKBuilder.app(SKBuilder.S, function.abstractPlusPlusPlusPlus(x), argument.abstractPlusPlusPlusPlus(x)) ;
    }

    public final boolean freeOf(SKSymbol x)
    {
	return function.freeOf(x) && argument.freeOf(x) ;
    }

    public void writeToWithArobase(Writer writer)
	throws Exception
    {
	//writeTo(writer) ;

	SKSymbol symbol = SKBuilder.getArobaseSymbol(this) ;

	if (symbol != null) {
	    writer.write("@") ;
	    symbol.writeTo(writer) ;
	    return ;
	}
	
	writeTo(writer) ;
    }
	    
    public void writeTo(Writer writer)
	throws Exception
    {
	function.writeToWithArobase(writer) ;
	writer.write(" ") ;
	if (argument.isAtom()) {
	    argument.writeTo(writer) ;
	} else {
	    writer.write("(") ;
	    argument.writeToWithArobase(writer) ;
	    writer.write(")") ;
	}
    }

    public boolean auxPushInStack()
    {
	SKMachine.pushRegisterAndSetTo(this) ;
	return true ;
    }

    public boolean operate() { /* never used */ return false ;  }

    public final SKTerm eval(int mode)
    {
	SKTerm result ;

	switch (mode) {
	case MachineMode.HEAD_REDUCTION:
	    SKMachine.reset(this) ;
	    SKMachine.executeHeadReduction() ;
	    return SKMachine.getResult() ;
	case MachineMode.FULL_REDUCTION:
	    SKMachine.reset(this) ;
	    SKMachine.executeHeadReduction() ;
	    (result = SKMachine.getResult()).evalFull() ;
	    return result ;
	case MachineMode.ONE_STEP_REDUCTION:
	    if ((result = evalOneStep()) != null)
		return result ;
	}

	return this ; // for a quiet compiler
    }

    public final void evalFull()
    {
	SKMachine.skAuxStack.push(this) ;
    }
	    
    public final SKTerm evalOneStep() 
    { 
	SKMachine.reset(this) ;

	if (SKMachine.executeOneStep())
	    return SKMachine.getResult() ;

	if (argument.isAtom())
	    return this ;

	SKMachine.reset(argument) ;

	if (SKMachine.executeOneStep()) {
	    argument = SKMachine.getResult() ;
	    return this ;
	}

	return null ; 
    } 
	    
    public final void replace(SKApplication app)
    {
	app.function = function ;
	app.argument = argument ;
    }

    public final SKTerm isCarAuxAux()
    {
	if (function.isK())
	    return argument ;

	return null ;
    }

    public final SKTerm isCarAux()
    {
	if (function.isSI())
	    return argument.isCarAuxAux() ;

	return null ;
    }

    public SKTerm isCar()
    {
	if (function.isS())
	    return argument.isCarAux() ;

	return null ;
    }

    public SKList isCdr()
    {
	if (function.isK()) {

	    SKList list = argument.isList() ;

	    if (list != null)
		return list ;

	    return new SKList(argument) ;
	}

	return null ;
    }

    public SKList isList()
    {
	if (function == SKBuilder.K && argument == SKBuilder.I)
	    return new SKList() ;

	SKList cdr = argument.isCdr() ;

	if (cdr != null) {

	    SKTerm car = function.isCar() ;

	    if (car != null) {
		cdr.add(car) ;
		return cdr ;
	    }
	}

	return null ;
    }

    public SKTerm interpret()
    {
	SKApplication copy = new SKApplication(function.interpret(),argument.interpret()) ;

	{
	    SKList list = copy.isList() ;

	    if (list != null)
		return list ;
	}

	if (copy.isNumber())
	    return SKBuilder.getSymbol(Integer.toString(copy.numberValue())) ;
	
	if (copy.isI())
	    return SKBuilder.I ;

	if (copy.isB())
	    return SKBuilder.getSymbol("B") ;

	if (copy.isW())
	    return SKBuilder.getSymbol("W") ;

	if (copy.isC())
	    return SKBuilder.getSymbol("C") ;

	// TBD

	return copy ;
    }

    public boolean isSuccessor()
    {
	return function.isS() && argument.isB() ;
    }

    public boolean isNumber()
    {
	return isKI() || (function.isSuccessor() && argument.isNumber()) ;
    }

    // Assume it is a number
    public int numberValue()
    {
	if (function.isSuccessor())
	    return 1 + argument.numberValue() ;
	return 0 ;
    }

    public boolean isB()
    {
	return function.isS1KS1() && argument.isK() ;
    }

    public boolean isS1KS1()
    {
	return function.isS() && argument.isKS() ;
    }

    public boolean isKS()
    {
	return function.isK() && argument.isS() ;
    }

    public boolean isI()
    {
	return function.isSK() && argument.isK() ;
    }

    public boolean isSK()
    {
	return function.isS() && argument.isK() ;
    }

    public boolean isW()
    {
	return function.isSS() && argument.isKI() ;
    }

    public boolean isSS()
    {
	return function.isS() && argument.isS() ;
    }

    public boolean isKI()
    {
	return function.isK() && argument.isI() ;
    }

    public boolean isC()
    {
	return function.isS1BBS1() && argument.isKK() ;
    }
    public boolean isKK()
    {
	return function.isK() && argument.isK() ;
    }

    public boolean isS1BBS1()
    {
	return function.isS() && argument.isBBS() ;
    }

    public boolean isBBS()
    {
	return function.isBB() && argument.isS() ;
    }

    public boolean isSI()
    {
	return function.isS() && argument.isI() ;
    }

    public final long nodeCount()
    {
	return 1L + function.nodeCount() + argument.nodeCount() ;
    }

}
