import java.awt.* ;
import java.awt.event.* ;
import javax.swing.* ;

public final class SKBuilder
{
    /* symbols */

    private static final SymbolTable symbolTable = new SymbolTable() ;

    public static final SKSymbol getSymbol(String name)
    {
	SKSymbol skSymbol = symbolTable.get(name) ;
	if (skSymbol != null)
	    return skSymbol ;
	skSymbol = new SKSymbol(name) ;
	symbolTable.put(name,skSymbol) ;
	return skSymbol ;
    }

    public static final void emptyWorkspace()
    {
	symbolTable.emptyWorkspace() ;
    }

    public static final void saveWorkspace(String workspaceFileName)
	throws Exception
    {
	symbolTable.saveWorkspace(workspaceFileName) ;
    }

    public final static void registerDefinition(SKSymbol s, SKTerm d)
    {
	s.setValue(d) ;
    }

    public final static SKArobase buildArobase(SKSymbol symbol)
    {
	return new SKArobase(symbol) ;
    }

    public final static void processArobases()
    {
	symbolTable.processArobases() ;
    }

    public final static SKSymbol getArobaseSymbol(SKTerm term)
    {
	return symbolTable.getArobaseSymbol(term) ;
    }

    /* inversed sequence to terms */

    public static final SKTerm buildApplication(SKSequence skSequence)
    {
	int last = skSequence.getLast() ;
	
	if (last == 0)
	    return null ;

	int    i      = last - 1 ;
	SKTerm result = skSequence.get(i) ;
	while (--i >= 0)
	    result = app(result,skSequence.get(i)) ;
	return result ;
    }

    /* applications */

    public static final SKApplication app(SKTerm p, SKTerm q, SKTerm... terms)
    {
	SKApplication result = new SKApplication(p,q) ;
	for (SKTerm term : terms)
	    result = new SKApplication(result,term) ;
	return result ;
    }

    /* S, K, I */

    public final static CombinatorS S = CombinatorS.unique ;
    public final static CombinatorK K = CombinatorK.unique ;
    public final static CombinatorI I = CombinatorI.unique ;

    /* B */

    public static final void setB() 
    { 
	SKTerm B = app(S,app(K,S),K) ;

	getSymbol("B").setValue(B) ; 
    }
    
    /* W */

    public static final void setW()
    { 
	SKTerm W = app(S,S,app(K,I)) ;

	getSymbol("W").setValue(W) ; 
    }

    /* C */
    
    public static final void setC()
    { 
	SKSymbol B = getSymbol("B") ;

	SKTerm C = app(S,app(B,B,S),app(K,K)) ;

	getSymbol("C").setValue(C) ;
    }

    /* C* */

    public static final void setCStar()
    { 
	SKSymbol C     = getSymbol("C") ;
	SKTerm CStar = app(C,I) ;

	getSymbol("C*").setValue(CStar) ;
    }

    /* Y */

    public static final void setY()
    {
	SKSymbol W = getSymbol("W") ;
	SKSymbol B = getSymbol("B") ;

	SKTerm Y = app(W,S,app(B,W,B)) ;

	getSymbol("Y").setValue(Y) ;
    }

    /* Kinf */

    public static final void setKinf()
    { 
	SKSymbol Y = getSymbol("Y") ;

	SKTerm Kinf = app(Y,K) ;

	getSymbol("K*").setValue(Kinf) ; 
    }

    /* omega */

    public static final void setOmega() 
    { 
	SKApplication omega = app(S,I,I) ;
	
	getSymbol("O").setValue(omega) ; 
    }

    /* OMEGA */

    public static final void setOMEGA() 
    { 
	SKSymbol omega = getSymbol("O") ;

	SKTerm OMEGA = app(omega,omega) ;

	getSymbol("OO").setValue(OMEGA) ; 
    }

    /* lambda + */

    public static final SKTerm lambdaPlus(SKSymbol x0, SKTerm m)
    {
	return m.abstractPlus(x0) ;
    }

    public static final SKTerm lambdaPlusPlus(SKSymbol x0, SKTerm m)
    {
	return m.abstractPlusPlus(x0) ;
    }

    public static final SKTerm lambdaPlusPlusPlus(SKSymbol x0, SKTerm m)
    {
	return m.abstractPlusPlusPlus(x0) ;
    }

    public static final SKTerm lambdaPlusPlusPlusPlus(SKSymbol x0, SKTerm m)
    {
	return m.abstractPlusPlusPlusPlus(x0) ;
    }

    public static final SKTerm lambdaPlus(SKSymbol x0, SKSymbol x1, SKTerm m)
    {
	return m.abstractPlus(x1).abstractPlus(x0) ;
    }

    public static final SKTerm lambdaPlus(SKSymbol x0, SKSymbol x1, SKSymbol x2, SKTerm m)
    {
	return m.abstractPlus(x2).abstractPlus(x1).abstractPlus(x0) ;
    }

    /* Q */

    public static final void setQ() 
    {
	SKSymbol x = getSymbol("x") ;

	SKTerm Q = lambdaPlus(x,app(x,K,S,K)) ;

	getSymbol("Q").setValue(Q) ;
    }

    /* true */

    public static final void setT() 
    { 
	SKTerm t = K ;

	getSymbol("TRUE").setValue(t) ; 
    }

    /* false */

    public static final void setF() 
    { 
	SKTerm f = app(K,I) ;

	getSymbol("FALSE").setValue(f) ; 
    }

    /* and */

    public static final void setAnd() 
    {
	SKSymbol C = getSymbol("C") ;
	SKSymbol f = getSymbol("FALSE") ;

	SKTerm and = app(app(C,C),f) ;

	getSymbol("AND").setValue(and) ; }

    /* or */

    public static final void setOr()
    {
	SKSymbol CStar = getSymbol("C*") ;
	SKSymbol t     = getSymbol("TRUE") ;

	SKTerm or = app(CStar,t) ;

	getSymbol("OR").setValue(or) ; 
    }

    /* not */

    public static final void setNot()
    { 
	SKSymbol CStar = getSymbol("C*") ;
	SKSymbol C     = getSymbol("C") ;
	SKSymbol f     = getSymbol("FALSE") ;
	SKSymbol t     = getSymbol("TRUE") ;

	SKTerm not = app(app(C, app(CStar,f)),t) ;

	getSymbol("NOT").setValue(not) ; 
    }

    /* cons */

    public static final void setCons() 
    { 
	SKSymbol a = getSymbol("a") ;
	SKSymbol b = getSymbol("b") ;
	SKSymbol x = getSymbol("x") ;

	SKTerm cons = lambdaPlus(a,b,x,app(x,a,b)) ;

	getSymbol("CONS").setValue(cons) ;
    }

    /* car */

    public static final void setCar() 
    {
	SKSymbol c = getSymbol("c") ;

	SKTerm car = lambdaPlus(c,app(c,K)) ;

	getSymbol("CAR").setValue(car) ; 
    }

    /* cdr */

    public static final void setCdr()
    { 
	SKSymbol c = getSymbol("c") ;

	SKTerm cdr = lambdaPlus(c,app(c,app(K,I))) ;

	getSymbol("CDR").setValue(cdr) ; 
    }

    /* nil */

    public static final void setNil()
    { 
	SKTerm nil = app(K,I) ;

	getSymbol("NIL").setValue(nil) ; }

    /* null */

    public static final void setNull()
    {
	SKSymbol CStar = getSymbol("C*") ;
	SKSymbol C     = getSymbol("C") ;
	SKSymbol t     = getSymbol("TRUE") ;
	SKSymbol f     = getSymbol("FALSE") ;

	SKTerm nullp = app(C,app(CStar,app(K,app(K,app(K,app(K,f))))),t) ;

	getSymbol("NULL").setValue(nullp) ; 
    }

    /* 0 et 1+ */

    public static final void setSucc() 
    { 
	SKSymbol B    = getSymbol("B") ;
	SKTerm succ = app(S,B) ;
	
	getSymbol("1+").setValue(succ) ;

	SKTerm zero = app(K,I) ;
	
	getSymbol("0").setValue(zero) ;
    }

    /* =0 */

    public static final void setZerop()
    {
	SKSymbol CStar = getSymbol("C*") ;
	SKSymbol C     = getSymbol("C") ;
	SKSymbol t     = getSymbol("TRUE") ;
	SKSymbol f     = getSymbol("FALSE") ;

	SKTerm zerop = app(C,app(CStar,app(K,f)),t) ;

	getSymbol("=0").setValue(zerop) ; 
    }

    /* 1- */

    public static final void setPred()
    { 
	SKSymbol n    = getSymbol("n") ;
	SKSymbol c    = getSymbol("c") ;
	SKSymbol cons   = getSymbol("CONS") ;
	SKSymbol car  = getSymbol("CAR") ;
	SKSymbol cdr  = getSymbol("CDR") ;
	SKSymbol zero = getSymbol("0") ;
	SKSymbol succ = getSymbol("1+") ;
	SKTerm pred = lambdaPlus(n,app(car,app(n,lambdaPlus(c,app(cons,app(cdr,c),app(succ,app(cdr,c)))),app(cons,zero,zero)))) ;

	getSymbol("1-").setValue(pred) ; }			     

    /* * */

    public static final void setTimes() 
    { 
	SKSymbol B     = getSymbol("B") ;
	SKSymbol C     = getSymbol("C") ;
	SKSymbol zero  = getSymbol("0") ;
	SKSymbol succ  = getSymbol("1+") ;
	SKTerm times = app(B,app(B,app(C,C,succ),B),app(C,C,zero)) ;

	getSymbol("*").setValue(times) ; 
    }

    /* + */

    public static final void setPlus() 
    {
	SKSymbol CStar  = getSymbol("C*") ;
	SKSymbol succ   = getSymbol("1+") ;

	SKTerm plus   = app(CStar,succ) ;

	getSymbol("+").setValue(plus) ;
    }

    /* - */

    public static final void setMinus()
    { 
	SKSymbol p     = getSymbol("p") ;
	SKSymbol q     = getSymbol("q") ;
	SKSymbol pred  = getSymbol("1-") ;

	SKTerm minus = lambdaPlus(p,q,app(q,pred,p)) ;

	getSymbol("-").setValue(minus) ; }

    /* <= */

    public static final void setLe()
    {
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol zerop = getSymbol("=0") ;
	SKSymbol minus = getSymbol("-") ;

	SKTerm le = lambdaPlus(x,y,app(zerop,app(minus,x,y))) ;

	getSymbol("<=").setValue(le) ; 
    }

    /* >= */

    public static final void setGe()
    {
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol zerop = getSymbol("=0") ;
	SKSymbol minus = getSymbol("-") ;
	
	SKTerm ge = lambdaPlus(x,y,app(zerop,app(minus,y,x))) ;
	
	getSymbol(">=").setValue(ge) ; 
    }

    /* < */

    public static final void setLt() 
    {
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol ge    = getSymbol(">=") ;
	SKSymbol not   = getSymbol("NOT") ;

	SKTerm lt = lambdaPlus(x,y,app(not,app(ge,x,y))) ;

	getSymbol("<").setValue(lt) ;
    }

    /* > */

    public static final void setGt() 
    {
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol le    = getSymbol("<=") ;
	SKSymbol not   = getSymbol("not") ;

	SKTerm gt = lambdaPlus(x,y,app(not,app(le,x,y))) ;

	getSymbol(">").setValue(gt) ;
    }

    /* == */

    public static final void setEq() 
    {
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol le    = getSymbol("<=") ;
	SKSymbol ge    = getSymbol(">=") ;
	SKSymbol and   = getSymbol("AND") ;

	SKTerm eq = lambdaPlus(x,y,app(and,app(le,x,y),app(ge,x,y))) ;

	getSymbol("==").setValue(eq) ;
    }

    /* != */

    public static final void setNeq() 
    {
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol lt    = getSymbol("<") ;
	SKSymbol gt    = getSymbol(">") ;
	SKSymbol or    = getSymbol("OR") ;

	SKTerm neq = lambdaPlus(x,y,app(or,app(lt,x,y),app(gt,x,y))) ;

	getSymbol("!=").setValue(neq) ;
    }

    /* / */

    public static final void setDiv()
    {
	SKSymbol f     = getSymbol("f") ;
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol Y     = getSymbol("Y") ;
	SKSymbol lt    = getSymbol("<") ;
	SKSymbol zero  = getSymbol("0") ;
	SKSymbol succ  = getSymbol("1+") ;
	SKSymbol minus = getSymbol("-") ;

	//SKTerm div = app(Y,lambdaPlus(f,lambdaPlus(x,y,app(app(lt,x,y),zero,app(succ,app(f,app(minus,x,y),y)))))) ;

	SKTerm div = lambdaPlus(x,y,app(app(lt,x,y),zero,app(succ,app(getSymbol("/"),app(minus,x,y),y)))) ;

	getSymbol("/").setValue(div) ; 
    }

    /* % */

    public static final void setMod()
    {
	SKSymbol f     = getSymbol("f") ;
	SKSymbol x     = getSymbol("x") ;
	SKSymbol y     = getSymbol("y") ;
	SKSymbol Y     = getSymbol("Y") ;
	SKSymbol lt    = getSymbol("<") ;
	SKSymbol minus = getSymbol("-") ;

	//SKTerm mod = app(Y,lambdaPlus(f,lambdaPlus(x,y,app(app(lt,x,y),x,app(f,app(minus,x,y),y))))) ;

	SKTerm   mod = lambdaPlus(x,y,app(app(lt,x,y),x,app(getSymbol("%"),app(minus,x,y),y))) ;

	getSymbol("%").setValue(mod) ; 
    }

    /* naturals */

    public static final SKTerm getNumber(int n)
    {
	SKTerm result =  getSymbol("0") ; 

	SKSymbol succ = getSymbol("1+") ;

	while (n-- > 0)
	    result = app(succ, result) ;
	return result ;
    }

    /* diverse cons */

    public final static SKApplication getCons(SKTerm a, SKTerm b)
    {
	return app(app(getSymbol("CONS"),a),b) ;
    }

    public final static SKApplication getCons(SKTerm a)
    {
	return getCons(a,getSymbol("NIL")) ;
    }

    public final static long nodeCount()
    {
	return symbolTable.nodeCount() ;
    }

}