import java.util.* ;
import java.awt.* ;
import java.awt.event.* ;
import javax.swing.* ;

public class SKCommandTextField extends JTextField
{
    private final SKCommandEditor skCommandEditor ;

    private final int    historySize = 1024 ;
    private final String history[] = new String[historySize] ;
    private       int    historyIdx = 0 ;
    private       int    historyCur = 0 ;

    public final void historyDown()
    {
	if (historyIdx >= historySize -1) {
	    SKUtils.beep() ;
	    return ;
	}

	history[historyIdx++] = getText() ;

	String str = history[historyIdx] ;
	setText(str) ;
	setCaretPosition(str.length()) ;

	historyCur = Math.max(historyCur,historyIdx) ;
    }    

    public final void historyUp()
    {
	if (historyIdx == 0) {
	    SKUtils.beep() ;
	    return ;
	}

	history[historyIdx--] = getText() ;

	String str = history[historyIdx] ;
	setText(str) ;
	setCaretPosition(str.length()) ;
    }

    public final void historyShift()
    {
	if (historyCur != historySize -1) {
	    history[historyCur++] = getText() ;
	    historyIdx = historyCur ;
	    return ;
	}

	for (int i = 1 ; i < historySize ; i++)
	    history[i-1] = history[i] ;

	history[historyCur] = getText() ;
	historyIdx = historyCur ;
    }

    public SKCommandTextField(SKCommandEditor _skCommandEditor)
    {
	this.skCommandEditor = _skCommandEditor ;

	Arrays.fill(history,"") ;

	setFont(SKUtils.monoFont) ;    

	addKeyListener
	    (new KeyAdapter() {

		    public void keyPressed(KeyEvent e) 
		    {
			switch (e.getKeyCode()) {

			case KeyEvent.VK_ENTER:
			    historyShift() ;
			    skCommandEditor.headReduction() ;
			    break ;

			case KeyEvent.VK_UP:
			case KeyEvent.VK_KP_UP:
			    historyUp() ;
			    break ;

			case KeyEvent.VK_DOWN:
			case KeyEvent.VK_KP_DOWN:
			    historyDown() ;
			    break ;

			case KeyEvent.VK_U:
			    if (e.isControlDown())
				historyUp() ;			
			    break ;

			case KeyEvent.VK_N:
			    if (e.isControlDown())
				historyDown() ;			
			    break ;
			}
		    }

		});
	
    }
}
