import java.io.* ;
import javax.swing.* ;

public final class SKMachine
{
    private static SKWorkerEvent eventRunning    = new SKWorkerEvent(SKWorkerEvent.FRAME_SET_RUNNING) ;
    private static SKWorkerEvent eventNotRunning = new SKWorkerEvent(SKWorkerEvent.FRAME_SET_NOT_RUNNING) ;

    public static void run(SKWorker skWorker, String text, int mode)
    {
	skWorker.skPublish(eventRunning) ;

	SKTerm term      = null ;
	long   startTime = 0L ;
	long   stopTime  = 0L ;

 	try {

 	    SKParser skParser = new SKParser(new StringReader(text)) ;

	    skParser.yyparse() ;

	    if (skParser.errorLine != -1)
		throw new Exception("Cannot decode a term") ;

	    term = skParser.expression ;

	    skParser = null ;

	    if (term == null)
		throw new Exception("Cannot decode a term") ;

	    term = term.elimArobases() ;

	    startTime     = System.currentTimeMillis() ;

	    SKTerm result = term.eval(mode) ;

	    while (mustFullContinue && skAuxStack.geThanOne()) {
		SKApplication app = skAuxStack.pop() ;
		app.argument = app.argument.eval(mode) ;
		app.function.evalFull() ;
	    }

	    stopTime      = System.currentTimeMillis() ;

	    if (!mustFullContinue)
		result = SKBuilder.getSymbol("#interrupted") ;

	    skWorker.skPublish(new SKWorkerEvent(SKWorkerEvent.DISPLAY_RESULT,stopTime - startTime,result,reductionCount)) ;
	    skWorker.skPublish(eventNotRunning) ;
	    
 	} catch (Throwable e) {
	    skWorker.setRunning(true) ;
	    skWorker.skPublish(new SKWorkerEvent(SKWorkerEvent.DISPLAY_ERROR, e.getMessage())) ;
	    skWorker.skPublish(eventNotRunning) ;
	    e.printStackTrace() ;
 	}

	System.gc() ;
    }

    public        static long          reductionCount ;
    public        static boolean       mustFullContinue ;
    public        static boolean       mustContinue ;
    private       static SKTerm        term ;
    public  final static SKStack       skStack    = new SKStack() ;
    public  final static SKStack       skAuxStack = new SKStack() ;
    public        static SKApplication skRegister ;

    public static void reset(SKTerm _term)
    {	
	mustContinue = mustFullContinue ;
	term         = _term ;
	skStack.reset() ;
    }

    public static final void pushRegisterAndSetTo(SKApplication app)
    {
	skStack.push(skRegister) ;
	skRegister = app ;
    }

    public static final void interrupt()
    {
	mustFullContinue = false ;
	mustContinue = false ;
    }

    public static final SKTerm getResult()
    {
	SKApplication result = skStack.getZero() ;

	if (result != null) {
	    if (SKMachine.skRegister.function == SKBuilder.I)
		SKMachine.skStack.top().function = SKMachine.skRegister.argument ;
	    return result ;
	}

	result = skRegister ;
	if (result.function == SKBuilder.I)
	    return result.argument ;

	return result ;
    }

    public static final boolean DEBUG = false ;

    public static final void executeHeadReduction()
    {
	skRegister = (SKApplication)term ;

	while (mustContinue) {

	    while (skRegister.function.auxPushInStack()) ;

	    if (DEBUG) {
		SKApplication tagada =  skStack.getZero() ;
		if (tagada != null) {
		    System.out.println("Eval: " + tagada.toString()) ; 
		} else {
		    System.out.println("Eval: null") ;
		}
		System.out.println("Register function: " + skRegister.function.toString()) ; 
 		System.out.println() ;
	    }

	    mustContinue = (mustContinue && skRegister.function.operate()) ;
	}
    }

    public static final boolean executeOneStep()
    {
	skRegister = (SKApplication)term ;

	while (skRegister.function.auxPushInStack()) ;

	return skRegister.function.operate() ;
    }

}
