import java.io.* ;
import java.awt.* ;
import java.awt.event.* ;
import javax.swing.* ;
import javax.swing.event.* ;
import javax.swing.filechooser.* ;

public final class SKProgramEditor extends JPanel
{
    private final SKFrame     skFrame ;
    private final JTextArea   skTextArea ;
    private final SKButton    compileProgramButton ;
    private final SKButton    compileSelectionButton ;
    private final SKButton    loadDefinedsButton ;
    private final SKLabelMono fileLabel ;
    private final SKLabelMono modifiedLabel ;
    private final SKLabelMono compiledLabel ;
    private final SKLabelMono lineLabel ;

    private       String      fileName = null;
    private       boolean     modified = false ;
    private       boolean     selected = false ;
    private       int         selectionDot ;
    private       int         selectionMark ;

    public SKProgramEditor(SKFrame skFrame)
    {
	setPreferredSize(new Dimension(999,400)) ;

	setBorder(BorderFactory.createTitledBorder("Program")) ;

	setLayout(new BorderLayout()) ;

	{
	    JPanel labelsPanel = new JPanel() ;

	    labelsPanel.add(fileLabel     = new SKLabelMono("[ no file ]", SwingConstants.LEFT),  BorderLayout.NORTH) ;
	    labelsPanel.add(modifiedLabel = new SKLabelMono("[ empty ]",   SwingConstants.RIGHT), BorderLayout.NORTH) ;
	    labelsPanel.add(compiledLabel = new SKLabelMono("[ empty ]",   SwingConstants.RIGHT), BorderLayout.NORTH) ;
	    labelsPanel.add(                new SKLabelMono("[ line ",     SwingConstants.RIGHT), BorderLayout.NORTH) ;
	    labelsPanel.add(lineLabel     = new SKLabelMono("0",           SwingConstants.RIGHT), BorderLayout.NORTH) ;
	    labelsPanel.add(                new SKLabelMono(" ]",          SwingConstants.RIGHT), BorderLayout.NORTH) ;

	    add(labelsPanel,BorderLayout.NORTH) ;
	}

	{
	    skTextArea = new JTextArea() ;
	    skTextArea.setFont(SKUtils.monoFont) ;
	    skTextArea.setLineWrap(true) ;
	    skTextArea.setWrapStyleWord(true);
	    skTextArea.setEditable(true) ;
	    skTextArea.getDocument().addDocumentListener(new DocumentListener()
		{
		    public void changedUpdate (DocumentEvent e) { textChanged() ; }
		    public void insertUpdate  (DocumentEvent e) { textChanged() ; }
		    public void removeUpdate  (DocumentEvent e) { textChanged() ; }
		}) ;
	    skTextArea.addCaretListener(new CaretListener()
		{
		    public void caretUpdate(CaretEvent e) 
		    {
			int dot  = e.getDot() ;
			int mark = e.getMark() ;
			selectionChanged(dot,mark) ;
		    }
		}) ;
	    JScrollPane jScrollPane = new JScrollPane(skTextArea) ;
	    jScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	    add(jScrollPane,BorderLayout.CENTER) ;
	}
	
	{
	    JPanel buttonsPanel = new JPanel() ;
	    buttonsPanel.setLayout(new GridLayout(1,4)) ;
	    buttonsPanel.add(new JLabel()) ;
	    buttonsPanel.add(compileProgramButton = new SKButton("Compile buffer",
								 new ActionListener() {
								     public void actionPerformed(ActionEvent e)
								     {
									 compileProgram() ;
								     }})) ;
	    buttonsPanel.add(compileSelectionButton = new SKButton("Compile selection",
								   new ActionListener() {
								       public void actionPerformed(ActionEvent e)
								       {
									   compileSelection() ;
								       }})) ;
	    buttonsPanel.add(loadDefinedsButton = new SKButton("Load predefineds",
								 new ActionListener() {
								     public void actionPerformed(ActionEvent e)
								     {
									 loadDefineds() ;
								     }})) ;
	    buttonsPanel.add(new JLabel()) ;
	    add(buttonsPanel,BorderLayout.SOUTH) ;
	}

	this.skFrame = skFrame ;
    }

    public final void selectionChanged(int selectionDot, int selectionMark)
    {
	selected = (selectionDot != selectionMark) ;

	if (selected) {
	    if (selectionDot < selectionMark) {
		this.selectionDot  = selectionDot ;
		this.selectionMark = selectionMark ;
	    } else {
		this.selectionDot  = selectionMark ;
		this.selectionMark = selectionDot ;
	    }
	}

	try {
	    lineLabel.setText(Integer.toString(skTextArea.getLineOfOffset(selectionDot))) ;
	} catch (Exception e) {} ;

	compileSelectionButton.setEnabled(selected) ;
    }

    public final void textChanged()
    {
	if (!modified) {
	    this.modifiedLabel.setText("[ not saved ]") ;
	    this.compiledLabel.setText("[ uncompiled ]") ;
	    modified = true ;
	}
    }

    public final void setRunning(boolean running)
    {
	boolean notRunning = !running ;

	skTextArea.setEnabled(notRunning) ;
	compileProgramButton.setEnabled(notRunning) ;
	compileSelectionButton.setEnabled(notRunning && selected) ;
	loadDefinedsButton.setEnabled(notRunning) ;
    }

    private final void clearBuffer()
    {
	if (modified) {
	    if (JOptionPane.showConfirmDialog(skFrame,
					      "Current edition buffer has not been saved and will be lost. Do you want to continue ?",
					      "Buffer not saved",
					      JOptionPane.OK_CANCEL_OPTION,
					      JOptionPane.WARNING_MESSAGE,
					      SKUtils.messageIcon) == JOptionPane.CANCEL_OPTION)
		return ;
	}

	skTextArea.setText("") ;

	this.modified = false ;
	this.fileLabel     .setText("[ no file ]") ;
	this.modifiedLabel .setText("[ empty ]") ;
	this.compiledLabel .setText("[ empty ]") ;
    }

    private JFileChooser skiFileChooser = null ;

    private final void createFileChooser()
    {
	skiFileChooser = new JFileChooser();
	skiFileChooser.setDialogTitle("Choose a .ski combinators program file") ;
	skiFileChooser.setCurrentDirectory(new File(".")) ;
	skiFileChooser.setFileView(new SKFileView()) ;
	FileNameExtensionFilter filter = new FileNameExtensionFilter("Combinators file","ski") ;
	skiFileChooser.setFileFilter(filter);
    }

    public final int quit()
    {
	if (modified) {
	    if (JOptionPane.showConfirmDialog(skFrame,
					      "Current edition buffer has not been saved and will be lost. Do you want to continue ?",
					      "Buffer not saved",
					      JOptionPane.OK_CANCEL_OPTION,
					      JOptionPane.WARNING_MESSAGE,
					      SKUtils.messageIcon) == JOptionPane.CANCEL_OPTION) {
		return 0 ; // means NO
	    } else {
		return 1 ; // means YES
	    }
	}
	return 2 ; // means NO_OPINION
    }

    public final void open()
    {
	if (modified) {
	    if (JOptionPane.showConfirmDialog(skFrame,
					      "Current edition buffer has not been saved and will be lost. Do you want to continue ?",
					      "Buffer not saved",
					      JOptionPane.OK_CANCEL_OPTION,
					      JOptionPane.WARNING_MESSAGE,
					      SKUtils.messageIcon) == JOptionPane.CANCEL_OPTION)
		return ;
	}

	if (skiFileChooser == null) 
	    createFileChooser() ;

	if (skiFileChooser.showOpenDialog(skFrame) !=  JFileChooser.APPROVE_OPTION)
	    return ;

	fileName = skiFileChooser.getSelectedFile().getName() ;

	try {
	    BufferedReader fileReader = new BufferedReader(new FileReader(fileName)) ;
	    String line ;

	    skTextArea.setText("") ;
	    do {
		line = fileReader.readLine() ;
		if (line != null) {
		    skTextArea.append(line) ;
		    skTextArea.append("\n") ;
		}
	    } while (line != null) ;

	    fileReader.close() ;

	    this.modified = false ;
	    this.fileLabel.setText("[ " + fileName + " ]") ;
	    this.modifiedLabel.setText("[ loaded ]") ;
	    this.compiledLabel.setText("[ uncompiled ]") ;

	} catch (Exception e) {
	    JOptionPane.showMessageDialog(skFrame,
					  "Error: " + e.getMessage(),
					  "Error when loading file in buffer",
					  JOptionPane.ERROR_MESSAGE,
					  SKUtils.messageIcon) ;
	}
    }

    private final void saveBuffer(String fileName)
    {
	try {
	    BufferedWriter fileWriter = new BufferedWriter(new FileWriter(fileName)) ;
	    fileWriter.write(skTextArea.getText()) ;
	    fileWriter.close() ;

	    this.fileName = fileName ;
	    this.modified = false ;
	    this.fileLabel.setText("[ " + fileName + " ]") ;
	    this.modifiedLabel.setText("[ saved ]") ;
	} catch (Exception e) {
	    JOptionPane.showMessageDialog(skFrame,
					  "Error: " + e.getMessage(),
					  "Error when saving buffer to file",
					  JOptionPane.ERROR_MESSAGE,
					  SKUtils.messageIcon) ;
	}
    }

    public final void save()
    {
	if (fileName != null) {
	    saveBuffer(fileName) ;
	} else {
	    saveAs() ;
	}
    }

    public final void saveAs()
    {
	if (skiFileChooser == null) 
	    createFileChooser() ;

	if (skiFileChooser.showSaveDialog(skFrame) !=  JFileChooser.APPROVE_OPTION)
	    return ;

	saveBuffer(skiFileChooser.getSelectedFile().getName()) ;
    }

    public final void emptyWorkspace()
    {
	if (JOptionPane.showConfirmDialog(skFrame,
					  "Erasing current definitions cannot be undone. Do you want to continue ?",
					  "Erasing definitions",
					  JOptionPane.OK_CANCEL_OPTION,
					  JOptionPane.WARNING_MESSAGE,
					  SKUtils.messageIcon) == JOptionPane.CANCEL_OPTION)
	    return ;

	SKBuilder.emptyWorkspace() ;
    }

    public final void saveWorkspace()
    {
	if (skiFileChooser == null) 
	    createFileChooser() ;

	if (skiFileChooser.showSaveDialog(skFrame) !=  JFileChooser.APPROVE_OPTION)
	    return ;

	try {
	    SKBuilder.saveWorkspace(skiFileChooser.getSelectedFile().getName()) ;
	} catch (Exception e) {
	    JOptionPane.showMessageDialog(skFrame,
					  "Error: " + e.getMessage(),
					  "Error when saving workspace to file",
					  JOptionPane.ERROR_MESSAGE,
					  SKUtils.messageIcon) ;
	}
    }

    private final void loadDefineds()
    {
	skFrame.loadDefineds() ;
    }

    private final void compileString(String text)
    {
 	try {
 	    SKParser skParser = new SKParser(new StringReader(text)) ;
	    skParser.yyparse() ;

	    int lineNumber = skParser.errorLine ;

	    if (lineNumber != -1) {
		if (!selected) {
		    int dot  = skTextArea.getLineOfOffset(lineNumber) ;
		    int mark = skTextArea.getLineEndOffset(lineNumber) ;
		    if (dot == mark) {
			skTextArea.setCaretPosition  (dot) ;
		    } else {
			skTextArea.setSelectionStart (dot) ;
			skTextArea.setSelectionEnd   (mark) ;
		    }
		}
		throw new Exception("Syntax error line " + lineNumber) ;
	    }
	    
	    SKBuilder.processArobases() ;

	    compiledLabel.setText("[ compiled ]") ;
	    
 	} catch (Exception e) {
	    JOptionPane.showMessageDialog(skFrame,
					  "Error: " + e.getMessage(),
					  "Compilation error",
					  JOptionPane.ERROR_MESSAGE,
					  SKUtils.messageIcon) ;
	    e.printStackTrace() ;
 	}
    }

    private final void compileProgram()
    {
	compileString(skTextArea.getText()) ;
    }

    private final void compileSelection()
    {
	compileString(skTextArea.getText().substring(selectionDot,selectionMark)) ;
    }

}

