import java.text.* ;
import java.io.* ;
import java.awt.* ;
import java.awt.event.* ;
import javax.swing.* ;

public final class SKResultDisplay extends JPanel
{
    private final SKFrame     skFrame ;
    private final JTextArea   skTextArea ;
    private final SKButton    headReductionButton ;
    private final SKButton    fullReductionButton ;
    private final SKButton    nextStepButton ;
    private final SKButton    interpretButton ;
    private final SKButton    interruptButton ;
    private final SKLabelMono skTimeDisplay ;
    
    public SKResultDisplay(SKFrame skFrame)
    {
	setPreferredSize(new Dimension(999,200)) ;

	setBorder(BorderFactory.createTitledBorder("Result")) ;

	setLayout(new BorderLayout()) ;

	add(skTimeDisplay = new SKLabelMono("[ ] [ ] [ ]",SwingConstants.CENTER),BorderLayout.NORTH) ;

	{
	    skTextArea = new JTextArea() ;
	    skTextArea.setFont(SKUtils.monoFont) ;
	    skTextArea.setLineWrap(true) ;
	    skTextArea.setWrapStyleWord(true);
	    skTextArea.setEditable(true) ;
	    
	    JScrollPane jScrollPane = new JScrollPane(skTextArea) ;
	    jScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	    add(jScrollPane,BorderLayout.CENTER) ;
	}
	
	{
	    JPanel buttonsPanel = new JPanel() ;
	    buttonsPanel.setLayout(new GridLayout(1,5)) ;

	    buttonsPanel.add(interpretButton = new SKButton("Interpret",
							    new ActionListener() {
								public void actionPerformed(ActionEvent e)
								{
								    interpret() ;
								}})) ;
	    buttonsPanel.add(headReductionButton = new SKButton("Head reduction",
								new ActionListener() {
								    public void actionPerformed(ActionEvent e)
								    {
									headReduction() ;
								    }})) ;
	    buttonsPanel.add(fullReductionButton = new SKButton("Full reduction",
								new ActionListener() {
								    public void actionPerformed(ActionEvent e)
								    {
									fullReduction() ;
								    }})) ;
	    buttonsPanel.add(nextStepButton = new SKButton("Next head redex",
							  new ActionListener() {
							      public void actionPerformed(ActionEvent e)
							      {
								  nextStepReduction() ;
							      }})) ;
	    buttonsPanel.add(interruptButton = new SKButton("Interrupt",
							    new ActionListener() {
								public void actionPerformed(ActionEvent e)
								{
								    SKMachine.interrupt() ;
								}})) ;

	    add(buttonsPanel,BorderLayout.SOUTH) ;
	}

	this.skFrame = skFrame ;
    }


    private final void headReduction()
    {
	SKWorker.run(skFrame,skTextArea.getText() + " ;;", MachineMode.HEAD_REDUCTION) ;
    }
	
    private final void fullReduction()
    {
	SKWorker.run(skFrame,skTextArea.getText() + " ;;", MachineMode.FULL_REDUCTION) ;
    }
	
    private final void nextStepReduction()
    {
	SKWorker.run(skFrame,skTextArea.getText() + " ;;", MachineMode.ONE_STEP_REDUCTION) ;
    }

    public final void setRunning(boolean running)
    {
	boolean notRunning = !running ;

	headReductionButton .setEnabled(notRunning) ;
	fullReductionButton .setEnabled(notRunning) ;
	nextStepButton      .setEnabled(notRunning) ;
	interpretButton     .setEnabled(notRunning) ;
	interruptButton     .setEnabled(running) ;
	skTextArea          .setEnabled(notRunning) ;
    }

    static public final DecimalFormat longFormatter 
	= new DecimalFormat("###,###,###,###,###.##") ;

    public final String millisToDuration(long duration)
    {
	if (duration == 0L)
	    return "0 ms" ;
	
	long ms = duration % 1000L ;
	
	String result = ms + " ms" ;
	
	duration = duration / 1000L ; // now sec
	
	if (duration == 0L)
	    return result ;
	
	long sec = duration % 60L ;
	
	result = sec + " sec " + result ;
	
	duration = duration / 60L ; // now min
	
	if (duration == 0L)
	    return result ;
	
	long min = duration % 60L ;
	
	result = min + " min " + result ;
	
	duration = duration / 60L ; // now hours
	
	if (duration == 0L)
	    return result ;
	
	long hours = duration % 24L ;
	
	result = hours + " hr " + result ;
	
	duration = duration / 24L ; // now days
	
	if (duration == 0L)
	    return result ;
	
	long days = duration ;
	
	result = days + " day(s) " + result ;
	
	return result ;
    }

    public final long speed(long duration, long reductionCount)
    {
	return new Double( ((double)duration) / (((double)reductionCount) / ((double)1000000))).longValue() ;
    }

    public final void displayResult(long duration, SKTerm term, long reductionCount)
    {
	skTimeDisplay.setText("[ " + millisToDuration(duration) + " ] "
			      + "[ " + longFormatter.format(reductionCount) + " red(s) ]"
			      + "[ " + speed(duration,reductionCount) + " ms/Mreds ]") ;
	displayResult(term) ;
    }
    
    private final void displayResult(SKTerm term)
    {
	skTextArea.setText(term.toString()) ;
    }

    public final void interpret()
    {
	SKTerm term = SKInterpret.interpret(skFrame,skTextArea.getText() + " ;;") ;

	if (term != null) {
	    displayResult(term) ;
	    return ;
	}

	JOptionPane.showMessageDialog(skFrame,
				      "Nothing to interpret !",
				      "Term interpretation",
				      JOptionPane.ERROR_MESSAGE,
				      SKUtils.messageIcon) ;
    }

}

