import java.io.* ;
import java.util.* ;

public final class SKStack
{
    private final int stackChunksIntialSize = 512 ;
    private final int stackChunksIncrement  = 128 ;

    private SKApplication stackChunks[][] ;
    private int           stackChunksSize ;
    private int           stackChunksLast ;
    private int           stackChunksIndx ;

    public final long longGetStackSize()
    {
	return stackChunksLast * chunkSize ;
    }

    public SKStack()
    {
	stackChunksSize = stackChunksIntialSize ;

	stackChunks = new SKApplication[stackChunksSize][] ;
	
	stackChunks[0]  = new SKApplication[chunkSize] ;
	stackChunksLast = 0 ;

	stackChunksIndx = 0 ;

	currChunk     = stackChunks[0] ;
	currChunkIndx = 0 ;

	reset() ;
    }

    private       SKApplication currChunk[]   ;
    private final int           chunkSize     = 16 * 1024 * 1024 ;
    private       int           currChunkIndx ;

    public final void reset()
    {
	while (notEmpty()) 
	    pop() ;
    }

    public final void push(SKApplication term)
    {
	if (currChunkIndx < chunkSize) {
	    currChunk[currChunkIndx++] = term ;
	    return ;
	}

	if (stackChunksIndx < stackChunksLast) {
	    currChunk     = stackChunks[stackChunksIndx++] ;
	    currChunkIndx = 0 ;
	    push(term) ;
	    return ;
	}

	if (stackChunksLast < stackChunksSize) {
	    stackChunks[stackChunksLast++] = new SKApplication[chunkSize] ;
	    push(term) ;
	    return ;
	}

	stackChunks = Arrays.copyOf(stackChunks,stackChunksSize += stackChunksIncrement) ;
	push(term) ;
	return ;
    }

    public final SKApplication getZero()
    {
	if (notEmpty())
	    return stackChunks[0][0] ;
	return null ;
    }

    public final boolean notEmpty() { return (currChunkIndx > 0) || (stackChunksIndx > 0) ; }

    /* Precondition: the stack is not empty */
    public final SKApplication pop()
    {
	SKApplication popped ;

	if (currChunkIndx > 0) {
	    currChunkIndx -= 1 ;
	    popped = currChunk[currChunkIndx] ;
	    currChunk[currChunkIndx] = null ;
	    return popped ;
	}

	currChunk     = stackChunks[--stackChunksIndx] ;
	currChunkIndx = chunkSize - 1 ;
	    
	popped = currChunk[currChunkIndx] ;
	currChunk[currChunkIndx] = null ;
	return popped ;
    }

    /* Precondition: the stack is not empty */
    public final SKApplication top() /* rank = 1, 2, ... */
    {
	if (currChunkIndx != 0)
	    return currChunk[currChunkIndx - 1] ;

	return stackChunks[stackChunksIndx - 1][chunkSize - 1] ;
    }

    /* Assuma rank < chunkSize, rank = 1 (top), 2, ... */
    public final SKApplication get(int rank) /* rank = 1, 2, ... */
    {
	if (rank <= currChunkIndx)
	    return currChunk[currChunkIndx - rank] ;

	if (stackChunksIndx > 0)
	    return stackChunks[stackChunksIndx - 1][chunkSize - (rank - currChunkIndx)] ;

	return null ;
    }

    public final boolean geThanOne() { return (currChunkIndx >= 1) || (stackChunksIndx > 0) ; }
    public final boolean geThanTwo() { return (currChunkIndx >= 2) || (stackChunksIndx > 0) ; }

    public final int getHeight() { return stackChunksIndx*chunkSize + currChunkIndx ; }
}


