import java.util.* ;

import java.io.* ;
import java.util.* ;
import java.lang.reflect.* ;

public class SymbolTable
{
    public final class KeyValueEntry
    {
	private final String   key ;
	private final SKSymbol value ;

	private KeyValueEntry next ;

	private KeyValueEntry(String key, SKSymbol value, KeyValueEntry next)
	{
	    this.key   = key ;
	    this.value = value ;
	    this.next  = next ;
	}

	public final String getKey() 
	{ 
	    return key ; 
	}

	public final SKSymbol getValue() 
	{ 
	    return value ; 
	}

	private final KeyValueEntry getNext() { return next ; }

	private final void setNext(KeyValueEntry next) { this.next = next ; }

    }

    private final KeyValueEntry[] table ;

    private final int innerSize ;
    private final int innerMask ;

    public SymbolTable()
    {
	this.innerSize = 512 ; /* Power of two ! */
	this.innerMask = innerSize - 1 ;
	this.table     = new KeyValueEntry[innerSize] ;
    }

    public final void put(String key, SKSymbol value)
    {
	int index = (key.hashCode() & innerMask) ;
	table[index] = new KeyValueEntry(key,value,table[index]) ;
    }

    public final SKSymbol get(String key)
    {
	KeyValueEntry entry = table[key.hashCode() & innerMask] ;

	while (entry != null) {
	    if (entry.getKey().equals(key))
		return entry.getValue() ;
	    entry = entry.getNext() ;
	}

	return null ;
    }

    public final void emptyWorkspace()
    {
	for (int i = 0 ; i < innerSize ; i++) {
	    KeyValueEntry entry = table[i] ;
	    while (entry != null) {
		entry.getValue().setValue(null) ;
		entry = entry.getNext() ;
	    }
	}
    }

    public final void saveWorkspace(String workspaceFileName)
	throws Exception
    {
	    BufferedWriter fileWriter = new BufferedWriter(new FileWriter(workspaceFileName)) ;

	    for (int i = 0 ; i < innerSize ; i++) {
		    
		KeyValueEntry entry = table[i] ;
		while (entry != null) {
		    entry.getValue().writeDefinition(fileWriter) ;
		    entry = entry.getNext() ;
		}
	    }

	    fileWriter.close() ;	
    }

    public final void processArobases()
    {
	for (int i = 0 ; i < innerSize ; i++) {
		    
	    KeyValueEntry entry = table[i] ;
	    while (entry != null) {
		entry.getValue().processArobases() ;
		entry = entry.getNext() ;
	    }
	}
    }

    public final SKSymbol getArobaseSymbol(SKTerm term)
    {
	for (int i = 0 ; i < innerSize ; i++) {
		    
	    KeyValueEntry entry = table[i] ;
	    while (entry != null) {
		SKSymbol symbol = entry.getValue() ;
		if (symbol.value == term)
		    return symbol ;
		entry = entry.getNext() ;
	    }
	}
	return null ;
    }

    public final long nodeCount()
    {
	int result = 0 ;

	for (int i = 0 ; i < innerSize ; i++) {
		    
	    KeyValueEntry entry = table[i] ;
	    while (entry != null) {
		SKSymbol symbol = entry.getValue() ;
		if (symbol.value != null)
		    result += symbol.value.nodeCount() ;
		entry = entry.getNext() ;
	    }
	}
	return result ;

    }


}
