import java.awt.event.* ;
import javax.swing.* ;

public final class SKFrame extends JFrame
{
    private final SKWindowPanel skWindowPanel ;
    private final SKMenuBar     skMenuBar ;

    public SKFrame()
    {
	super("SK Machine - Proof of concept") ;

	setJMenuBar(skMenuBar = new SKMenuBar(this)) ;

	setContentPane(skWindowPanel = new SKWindowPanel(this)) ;

	setRunning(false) ;

	addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) 
		{
		    quit() ;
 		}
 	    });
	
	setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

	pack() ;           
	setVisible(true) ;  
    }

    public final void setRunning(boolean running)
    {
	skMenuBar.setRunning(running) ;
	skWindowPanel.setRunning(running) ;
    }

    public final void loadDefineds()
    {
	skWindowPanel.loadDefineds() ;
    }

    public final void about()
    {
	JOptionPane.showMessageDialog(this,
				      "JSKI\n" +
				      "An implementation of Turner's SK Machine.\n" +
				      "Patrick Bellot, Telecom ParisTech, 2011.",
				      "About JSKI",
				      JOptionPane.INFORMATION_MESSAGE,
				      SKUtils.messageIcon) ;
    }

    public final void open()
    {
	skWindowPanel.open() ;
    }

    public final void save()
    {
	skWindowPanel.save() ;
    }

    public final void saveAs()
    {
	skWindowPanel.saveAs() ;
    }

    public final void quit()
    {
	switch (skWindowPanel.quit()) {
	case 0: // means NO
	    return ;
	case 1: // means YES
	    System.exit(0) ;
	    break ;
	default: // means no opinion
	    break ;
	}

	if (JOptionPane.showConfirmDialog(this,
					  "Really quit ?",
					  "SKI quit",
					  JOptionPane.OK_CANCEL_OPTION,
					  JOptionPane.WARNING_MESSAGE,
					  SKUtils.messageIcon) == JOptionPane.CANCEL_OPTION)
	    return ;
	System.exit(0) ;
    }

    public final void emptyWorkspace()
    {
	skWindowPanel.emptyWorkspace() ;
    }

    public final void saveWorkspace()
    {
	skWindowPanel.saveWorkspace() ;
    }

    public final void displayResult(long duration, SKTerm term, long reductionCount)
    {
	skWindowPanel.displayResult(duration,term,reductionCount) ;
    }

}
